comment|   FOR QB/PDS NEAR STRINGS ONLY!    Compile using MASM 5.1

  Written by Brian McLaughlin, released into public domain 11/21/93

  There are four routines in this module. The main one (BufInput$) is
  a routine to replace LINE INPUT# for reading in ASCII text files
  one line at a time.  It's MUCH FASTER than LINE INPUT# because it
  uses a buffer of 32K (by default) to minimize the number of file
  reads (the slowest part of file input).

  BufInput$ is designed to be called in a loop. At each iteration it
  returns the next line of text, as delimited by CHR$(13) + CHR$(10),
  the carriage return/line feed characters in ASCII. It is designed to
  ignore ^Z, end-of-file markers. If one is appended to the end of the
  file being read, BufInput$ clips it off. This may not be 100% the
  way other people would design things, but I like it that way.

  BufInput$ requires the DOS handle number of the file to read, NOT the
  BASIC file number. You can obtain the handle using FILEATTR like so:

        DOSHandle% = FILEATTR(FileNum%, 2)

  The other routines in the module are BufEOF%, BufError% and BufSetSize.
  BufEOF% acts just like EOF -- it returns 0 until the end of the file is
  reached, in which case it returns -1.

  Whenever BufInput$ encounters a DOS error, then the DOS error number can
  be obtained by your program by calling BufError%. When there is no error,
  BufError% returns 0.

  BufInput$ defaults to a buffer size of 32K. It acquires the memory by
  making a call to SETMEM (a QB function) first -- to release 32K from
  QB's memory management, then it allocates the buffer by calling upon DOS.
  When the end of a file is reached, the buffer memory is returned to QB.

  If for some reason QB cannot release a full 32K, you may wish to resize
  the buffer using BufSetSize. BufSetSize doesn't allow you to set a buffer
  size smaller than 1K. To find out how much memory QB can spare, do this:

        MemAvailable& = SETMEM(0)
        IF MemAvailable& < 32768 THEN
           NewSize% = CINT(MemAvailable&) - 32
           BufSetSize(NewSize%)
        END IF

  Put the following declarations in your code:

  DECLARE FUNCTION BufInput$ (FileHandle%)
  DECLARE FUNCTION BufEOF% ()
  DECLARE FUNCTION BufError% ()
  DECLARE SUB BufSetSize (BytesInBuffer%)

end comment|

.MODEL MEDIUM
    Extrn        B$SETM:PROC
.DATA
    Len          DW ?
    Addr         DW ?
    StrData      DB 128 DUP(0)
.CODE
    BufHandle    DW 0FFFFh ;set initial handle to -1
    BufSeg       DW 0      ;segment in far heap allocated to the buffer
    BufSize      DW 8000h  ;set initial buffer size at 32K
    BufFilled    DW 0      ;number of bytes read into buffer most recently
    BufIndex     DW 0      ;index to byte in buffer where scan starts
    RemainsLow   DW 0      ;number of bytes remaining to be read into buffer
    RemainsHigh  DW 0
    TotalLow     DW 0      ;saves the total size of file in bytes
    TotalHigh    DW 0
    Error        DW 0      ;saves any error code from latest operation
    EndFile      DW 0      ;flag set to -1 when end of file reached

PUBLIC BufSetSize, BufInput, BufError, BufEOF


BufError PROC FAR
    Mov AX, CS:Error
    Retf
BufError ENDP


BufEOF PROC FAR
    Mov AX, CS:EndFile
    Retf
BufEOF ENDP

BufInput PROC FAR           ; called with Handle% in [BP+6]
    Push BP
    Mov BP, SP
    Push DS
    Push ES
    Push DI
    Push SI
    Mov AX, Offset StrData
    Mov Addr, AX           ;build a string descriptor
    Mov Len, 0             ;assume a null string will be returned
    Mov CS:Error, 0        ;assume no errors, clear any old error code
    Mov BX, [BP+6]         ; BX = Handle% address
    Mov BX, [BX]           ;BX = Handle% value
    Cmp BX, 5              ; check if Handle% is a DOS reserved handle
    Jb  JmpToEOF           ; if so, jump to intermediate label
    Cmp BX, CS:BufHandle   ;is this handle one we're already reading?
    Jne NewHandle          ;if it's a new handle, go initialize buffer, etc.
    Mov AX, CS:BufSeg      ;IMPORTANT: DS points to buffer in far heap!!
    Mov DS, AX             ; DS does NOT point to the data segment now
    Mov ES, AX             ; point ES at buffer! (DS points there, too)
Initialized:
    Mov CX, CS:BufFilled   ; CX = total bytes read into buffer by $FillBuffer
    Jcxz JmpToEOF          ;if there's no bytes in buffer, return a null$
    Mov DI, CS:BufIndex    ;ES:DI point to starting byte of next scan
    Sub CX, DI             ;CX = bytes left to scan in buffer
    Cmp CX, 128            ; are there more than 128 bytes left to scan?
    Jb  NearBufEnd         ; if so, stop and refill the buffer, if necessary
    Mov CX, 128            ; don't scan more than 128 bytes at a time
LessThan128:
    Mov DX, CX             ; save the initial value of CX in DX (for later)
    Mov AL, 13             ;scan for a carriage return char - CHR$(13)
    Xor BX, BX
    Repne Scasb            ;repeat CX times or until there's a match
    Jnz NoCRFound          ;if NO CHR$(13), zero flag is clear (and CX=0)
    Sub DX, CX             ;subtract bytes beyond CHR$(13) from bytes scanned
    Dec DX                 ;DX = num of bytes from BufIndex to first CHR$(13)
    Add BX, 2              ; adjusts BufIndex for one CR/LF pair
NoCRFound:
    Mov SI, Offset Len     ; SI = the offset of Len in data seg
    Mov SS:[SI], DX        ; put DX in Len: string descriptor is complete
    Call $StringFill       ; this routine fills StrData with return string
    Add CS:BufIndex, BX    ; BX = 0 or 2, depending on if CR/LF found
    Add CS:BufIndex, DX    ; BufIndex points to char to start next scan with
    Mov CX, CS:BufFilled   ; calculate if we're at the last byte or none
    Sub CX, CS:BufIndex
    Cmp CX, 2
    Jb  TestLastChar
BufExit:
    Mov AX, Offset Len
    Pop SI
    Pop DI
    Pop ES
    Pop DS
    Pop BP
    Retf 2
JmpToEOF:
    Jmp SHORT SetEOF
NewHandle:
    Call $InitHandle       ; initialize a new buffer & internal variables
    Jc  BufExit            ; if carry set, $InitHandle had an error
    Mov AX, CS:BufSeg      ;IMPORTANT: DS points to buffer in far heap!!
    Mov DS, AX
    Mov ES, AX             ; ES points to buffer now, too! (like DS does)
    Call $FillBuffer       ; fills buffer, resets BufIndex, BufFilled
    Jc  BufExit            ; if carry set, $FillBuffer had an error
    Jmp SHORT Initialized  ;go back to scan buffer for first return string
NearBufEnd:
    Mov AX, CS:BufFilled   ; was the last read < BufSize?
    Cmp AX, CS:BufSize     ; if they're NOT equal, then NO more file to read
    Jne LessThan128        ;if nothing left to read in, we're near end
    Mov AX, CS:RemainsLow
    Or  AX, AX
    Jnz BufRefill          ; if RemainsLow > 0, we need to refill buffer
    Mov AX, CS:RemainsHigh ; if low byte = 0, we need to look at the high byte
    Or  AX, AX
    Jz  LessThan128        ; this falls through if high byte > 0
BufRefill:
    Call $ResetFilePtr     ; if we're here, we need to refill buffer
    Jc  BufExit            ; carry flag set if $ResetFilePtr had an error
    Call $FillBuffer       ; fills buffer, resets BufIndex, BufFilled
    Jc  BufExit            ; if carry set, $FillBuffer had an error
    Jmp Initialized        ; jump (long) to start over with new values
TestLastChar:
    Mov SI, BufFilled
    Cmp Byte Ptr [SI-1], 26
    Je  RemoveEOFMarker
    Jcxz SetEOF            ;if no bytes left, just return StrData as is
    Jmp SHORT BufExit      ;there's one byte left, it isn't CHR$(26)
RemoveEOFMarker:
    Mov SI, Offset Len
    Mov AX, SS:[SI]
    Dec AX
    Mov SS:[SI], AX        ;this line falls through, past SetEOF label
SetEOF:
    Cmp CS:BufSeg, 0
    Je  Released
    Call $ReleaseMem       ;give buffer memory back to QB
    Jc  BufExit
Released:
    Mov CS:EndFile, 0FFFFh    ;set the EOF marker = (-1)
    Mov CS:BufHandle, 0FFFFh  ;set "current" handle to impossible number (-1)
    Mov CS:BufSeg, 0
    Jmp BufExit
BufInput ENDP

BufSetSize PROC FAR  ;lets user set a new buffer size between 1K and 32K
    Push BP                ; if Bytes% is >32K, BufSetSize adjusts it to 32K
    Mov BP, SP             ; (if Bytes% not evenly divisible by 16, procedure
    Mov BX, [BP+6]         ; will round Bytes% up to the next multiple of 16)
    Mov AX, [BX]           ;AX = Bytes%
    Xor BX, BX             ; set flag in BX (assumes Bytes% MOD 16 = 0)
    Cmp AX, 1024           ;see if Bytes% is less than 1024 bytes
    Jb  AdjustUp           ; if so, go adjust it up to 1024 bytes
    Test AX, 15            ;see if (Bytes% MOD 16) = 0
    Jnz RoundUp            ;if not, we'll adjust up to nearest multiple of 16
SetResume1:
    Mov CL, 4
    Shr AX, CL             ;convert Bytes% to paragraphs (Bytes%\16)
    Add AX, BX             ;BX = 1 or 0, depending if (Bytes% MOD 16) = 0
    Cmp AX, 2048
    Ja  AdjustDown         ; when Bytes% is negative buffer size = 32K
SetResume2:
    Mov CL, 4
    Shl AX, CL             ; convert paragraphs back to bytes
    Mov CS:BufSize, AX     ; save it
SetExit:
    Pop BP
    Retf 2
RoundUp:
    Inc BX
    Jmp SHORT SetResume1
AdjustUp:
    Mov AX, 64             ;don't allow buffer size < 1K
    Jmp SHORT SetResume2
AdjustDown:
    Mov AX, 2048           ;don't allow buffer size > 32K
    Jmp SHORT SetResume2
BufSetSize ENDP

; NOTE: all the following procedures (with names beginning with $) are
; internal routines, called from within BufInput PROC.

$StringFill PROC NEAR  ; copies Return$ into StrData: trashes ES, SI, DI
    Push SS                ; point ES:DI at StrData (ES = data seg)
    Pop ES                 ;NOTE: DX holds length of return string on entry
    Mov CX, DX             ;CX = length of return string
    Jcxz StrExit           ; if null, we can just exit now
    Mov DI, Offset StrData
    Mov SI, CS:BufIndex    ; and point DS:SI at start byte of last scan
    Rep Movsb              ; copy CX bytes into StrData
StrExit:
    Retn
$StringFill ENDP

$ResetFilePtr PROC NEAR    ;reset pointer before next buffer fill
    Add CS:RemainsLow, CX  ;add bytes left in buffer to the bytes remaining
    Adc CS:RemainsHigh, 0  ;and make sure any extra carries into high word
    Mov DX, CS:TotalLow
    Mov CX, CS:TotalHigh
    Sub DX, CS:RemainsLow  ;put desired location into CX:DX
    Sbb CX, CS:RemainsHigh ;as an offset from the start of file
    Mov BX, CS:BufHandle   ;DOS needs the handle in BX
    Mov AX, 4200h          ;use offset-from-end service
    Int 21h                ;do it
    Jc  ResetError         ;carry flag set if error
ResetExit:
    Retn
ResetError:
    Mov CS:Error, AX
    Jmp SHORT ResetExit
$ResetFilePtr ENDP


$FillBuffer PROC NEAR   ; trashes AX,BX,CX,DX,DI,SI & assumes DS = BufSeg
    Mov BX, CS:BufHandle   ; we need the handle in BX for DOS call
    Xor DX, DX             ;point DS:DX at buffer's first byte
    Mov CS:BufIndex, DX    ;whenever buffer is filled index set to 0
    Mov CX, CS:BufSize     ;fill the buffer with CX bytes from file
    Mov AX, 3F00h          ;request read-file service
    Int 21h                ; from DOS -- returns bytes read in AX
    Jc  FillBufError       ; if no carry flag, the buffer is filled
    Sub CS:RemainsLow, AX  ;show there are AX fewer bytes left in file
    Sbb CS:RemainsHigh, 0  ;make certain high word also reflects the subtract
    Cmp AX, CX             ;did we read an entire block, BufSize bytes long?
    Je  FillBufExit        ; $InitHandle set BufFilled = BufSize
    Mov CS:BufFilled, AX   ; if not, save actual BufFilled
    Clc
FillBufExit:
    Retn
FillBufError:
    Mov CS:Error, AX       ; save the error
    Jmp SHORT FillBufExit
$FillBuffer ENDP

$ReleaseMem PROC NEAR
    Mov AX, CS:BufSeg
    Mov ES, AX             ;set ES to segment of buffer
    Mov AX, 4900h          ;DOS release-memory service
    Int 21h
    Jc  ReleaseError
    Mov AX, 7FFFh
    Push SS
    Pop DS                 ;SETMEM needs DS = data seg
    Push AX
    Push AX                ;tell SETMEM to grab back deallocated memory
    Call B$SETM
    Clc                    ;explicitly clear carry flag that SETMEM set
ReleaseExit:
    Retn
ReleaseError:
    Mov CS:Error, AX
    Jmp SHORT ReleaseExit
$ReleaseMem ENDP

$InitHandle PROC NEAR
    Mov CS:BufHandle, BX   ; save handle as current handle
    Mov CS:EndFile, 0      ; zero this variable when there's a new file
    Xor DX, DX
    Xor CX, CX
    Mov AX, 4202h          ; move zero bytes from end-of-file
    Int 21h                ; DOS returns file size in DX:AX
    Jc  InitError
    Mov CS:RemainsLow, AX  ;save the size of the file (DX:AX) in bytes
    Mov CS:RemainsHigh, DX
    Mov CS:TotalLow, AX
    Mov CS:TotalHigh, DX
    Xor DX, DX
    Mov AX, 4200h          ; move zero bytes from first byte of file
    Int 21h
    Jc  InitError
    Mov BX, CS:BufSize     ;BX = length of buffer in bytes
    Mov CS:BufFilled, BX   ; initialize BufFilled = BufSize, adjust later
    Mov CL, 4
    Shr BX, CL             ; convert bytes to paragraphs (shift-divide by 16)
    Mov AX, BX             ; note: BX holds number of paragraphs to allocate
    Inc AX                 ;SETMEM is peculiar, so this is required
    Mov CL, 4
    Shl AX, CL             ; convert paragraphs back to bytes (in AX)
    Xor DX, DX
    Neg AX                 ; 32-bit negate preserves the sign in DX:AX, so...
    Sbb DX, 0              ;DX:AX holds bytes-to-release as a negative number
    Push DX
    Push AX                ; leave it on the stack for SETMEM
    Call B$SETM            ; SETMEM releases the memory (and trashes DX, AX)
    Mov AX, 4800h          ; request DOS allocate memory service
    Int 21h                ; it requires paragraphs-to-allocate in BX
    Jc  InitError          ; if carry set, there's an error
    Mov CS:BufSeg, AX      ; otherwise, AX holds the segment for buffer
InitExit:
    Retn
InitError:
    Mov CS:Error, AX       ; save the error
    Jmp SHORT InitExit
$InitHandle ENDP

    END
